( * 4x4 matrix keypad support for 3-instruction Forth - Assad Ebrahim - 2020-07-25---Sat---21:06 )

8 constant  CALL_KEYPAD     \ poll keypad (in 3iKeypad.fs OR 3iDrumPad.fs)

\ 4x4 matrix keypad - 8 pins
\ TX lines, white tabbed, first four counting from left with pins facing up
2  constant padtx1 
3  constant padtx2 
4  constant padtx3 
5  constant padtx4 
\ RX lines, black tabbed, last four counting from the left with pins facing up
6  constant padrx1
7  constant padrx2
8  constant padrx3
9  constant padrx4

\ ========================================================

( * 4x4 Matrix Keypad - Membrane or Button * )
( * How it works: )
(     4 pins are tx/output, 4 are rx/input.  )
(     To figure out if/which key was pressed, uC flips one tx pin live at a time, and polls the rx pins )
(     A pressed button will mean we have row/col and logic determines which key was pressed )

: keypad-init  \ doesn't matter which 4 are input and which 4 are output --- it's symmetric
	padtx1 out padtx2 out padtx3 out padtx4 out
	padrx1 pullup  padrx2 pullup  padrx3 pullup  padrx4 pullup
;

\  !!!!!! ONLY NEEDED DURING DEVELOPMENT OF 4x4 KEYPAD --- with pinread() -- re-use for other polled sensors
\ CONSIDER target routine for write nibble, read nibble
\ CONSIDER use 74165 shift-in register IC to reduce pin count for controlling these keypads from 8 to 3
\ Application: Can map the keys to anything - output or instruction set - e.g. a macrokeypad via Forth...
: poll-keypad ( -- )
	padtx1 low
	padrx1 pinread 1 - if ." S4 " then  
	padrx2 pinread 1 - if ." S3 " then  
	padrx3 pinread 1 - if ." S2 " then  
	padrx4 pinread 1 - if ." S1 " then
	padtx1 high
	
	padtx2 low
	padrx1 pinread 1 - if ." S8 " then  
	padrx2 pinread 1 - if ." S7 " then  
	padrx3 pinread 1 - if ." S6 " then  
	padrx4 pinread 1 - if ." S5 " then
	padtx2 high
	
	padtx3 low
	padrx1 pinread 1 - if ." S12 " then  
	padrx2 pinread 1 - if ." S11 " then  
	padrx3 pinread 1 - if ." S10 " then  
	padrx4 pinread 1 - if ." S9 "   then
	padtx3 high
	
	padtx4 low
	padrx1 pinread 1 - if ." S16 " then  
	padrx2 pinread 1 - if ." S15 " then  
	padrx3 pinread 1 - if ." S14 " then  
	padrx4 pinread 1 - if ." S13 "  then	
	padtx4 high
;

: keypad-test begin poll-keypad key? until ;

\ MATCHED with 3iKeypad.avr
create keypad-chars '0' c, '1' c, '2' c, '3' c, '4' c, '5' c, '6' c, '7' c, '8' c, '9' c, 'A' c, 'B' c, 'C' c, 'D' c, '#' c, '*' c,
create keypad-nums  13 c, 0 c, 1 c, 2 c, 4 c, 5 c, 6 c, 8 c, 9 c, 10 c, 3 c, 7 c, 11 c, 15 c, 14 c, 12 c,

\ DESIGN: key press flag signals through the layers of the loop when to stop polling
variable pollstate
: pollstate! ( n -- ) pollstate ! ;
: pollstate@ ( -- n ) pollstate @ ;
: pollstop  1 pollstate! ;
: resetpoll 0 pollstate! ;

resetpoll

: keypadread ( -- val )  \ read matrix membrane keypad and send back 0x0-0xF numbers
	CALL_KEYPAD xcall
	100 ms  \ 50ms works but you get multiple reports per press
	1 [read-com if rcv-buffer c@ else $FF then ;

: fastkeypad ( -- key )  \ software polling --- could do other things to interrupt/do other instructions
    begin 
        keypadread dup $FF = if drop 0 else 1 then 
        key? if pollstop then pollstate@ or 
    until ;

: fastkeypad-chars ( -- )  \ emits which keys are pressed, as ASCII characters not numbers
	 cr resetpoll begin fastkeypad pollstate@ if -1 else bytes keypad-chars + c@ emit space 0 then until ;

: fastkeypad-num ( -- n )  \ emits one number onto stack
	 cr resetpoll fastkeypad pollstate@ if -1 else bytes keypad-nums + c@ dup . space then ;

: fastkeypad-nums ( -- nums onto stack )  \ emits numbers pressed onto stack
	 cr resetpoll begin fastkeypad pollstate@ if -1 else bytes keypad-nums + c@ dup . space 0 then until ;

\ TODO: implement old-style phone letter selection... if second press within timeout, switch the letter, otherwise accept


\ ====== APPLICATION  (include 3iBeats.fs)
\  KEYPAD TEST -- include 3iKeypad.fs
 \  : beep ( -- ) buzzer hi 5 ms buzzer lo 50 ms ;
 \  : buzzcount ( N -- ) 0 do beep loop ;
 \  : keypad-demo ( -- ) 
 \      begin  fastkeypad 1 + buzzcount  key?  until key drop ;
        
\ APPLICATION - use 74165 to allow a simple ATTINY to monitor a keypad continuously and send out the results, so not to slow down the main uC.  Another application for the GA144...

 \  create beatspad 0 c, 1 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 0 c, 
 \  : drawnote ( 1/0 -- ) dup 1 = if 14 else 22 then emit space ;
 \  : nibblespace? ( -- ) 1 + 4 mod 0 = if 2 spaces then ;  \ put space after each 4 bits (nibble) to aid visual perception
 \  : beats-demo ( -- )
 \  cr ." Enter a quartet of 16th-note beats using hex digits on keypad: " cr
 \      4 0 do fastkeypad bytes beatspad + c@ drawnote i nibblespace? loop >b4 1 Nseq-nibble-msb  \ get 8 entries, then loop it
 \  ; 

\ ======== KEYPAD PORTION =========
\ keypad-init
\  : beat-keypad ( -- )  \ input rhythm from 4x4 matrix keypad, 1 quarter note (4 beats) at a time.
\                     \ ex: 8 8 11 0   10 11 0 0   8 8 11 0   10 15 8 0
\      reset-stop resetbytes draw-menu
\      cr ." Input beats (hex digits)." cr
\      begin fastkeypad-num bytecount++ bytecount@ nseq-nibble-msb key? until key drop
\  ;
\ TODO: wrap into menu...
\ _________________________________


\ ====== APPLICATION \ ======== 5 button menu
( BUTTON LAYOUT:
    k3-play k2-clr k1-rest
     k4-clrall k5-beat
)

\ 5-button menu controls use Nano pins 2-6
\ 2 constant k1-rest
\ 3 constant k2-clr
\ 4 constant k3-play
\ 5 constant k4-clrall
\ 6 constant k5-beat
\ 
\ : buttons-init
\     k1-rest pullup \ buttons connect pin to ground and are active low
\     k2-clr pullup
\     k3-play pullup
\     k4-clrall pullup
\     k5-beat pullup
\ ;




\ =================================================

: help-keypad
  cr
  cr  ." KEYPAD (help-keypad)"
  cr 
  cr  ."  Firmware: 3iKeypad.avr (NESTED2) 3iI2C.avr  3iForth.avr"
  cr  ." include 3iForth.fs  3iKeypad.fs"
  cr  ." 1 open-com   rx"
  cr
  cr  ." 4X4 MATRIX KEYPAD"
  cr  ." keypad-init   create keypad (16 values)"
  cr  ." fastkeypad   fastkeypad-test"
  cr
;

help-keypad
